/*
 * @(#)JOTTOEditor.java 1.19 01/12/03
 *
 * Copyright 2002 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package Editor;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.io.*;
import java.net.URL;
import java.net.MalformedURLException;
/*import java.net.URI;
import java.net.URISyntaxException;*/
import java.util.*;

import javax.swing.text.*;
import javax.swing.undo.*;
import javax.swing.event.*;
import javax.swing.*;



/**
 * Sample application using the simple text editor component that
 * supports only one font.
 *
 * @author  Timothy Prinzing
 * @version 1.19 12/03/01
 */
class JOTTOEditor extends JPanel 
{
    
    private static boolean isSourcePos(String selText, int[]pos)
    {
        StringTokenizer st=new StringTokenizer(selText,".");
        try
        {
            int i;
            for(i=0;st.hasMoreTokens();i++)
            {
                
                pos[i]=Integer.parseInt(st.nextToken());
            }
            if(i>2)
                return false;
            return true;
        }
        catch(Exception e)
        {   
        }
        return false;
    }

    private static ResourceBundle resources;
    private String currentDocumentFileName;
    private String currentDocumentFileDirectory;
    private String outputDirectory = ".\\";
    private String outputInfoDirectory = ".\\";
    private String outputOBJDirectory = ".\\";
    private String outputEXEDirectory = ".\\";

    private final Properties    properties         = new Properties();
    private final static String settingsProperties = "Editor/resources/settings.properties";
    private final File          propertiesFile     = new File( settingsProperties );

    static 
    {
        try 
        {
            resources = ResourceBundle.getBundle("Editor.resources.JOTTOEditor",
                                                 Locale.getDefault());
        } catch (MissingResourceException mre) 
        {
            System.err.println("resources/JOTTOEditor.properties not found");
            System.exit(1);
        }
    }

    JOTTOEditor() 
    {
        super(true);

        // Force SwingSet to come up in the Cross Platform L&F

        /*System.out.println( "java.class.path = " + System.getProperty( "java.class.path" ) );
        String path = System.getProperty( "java.class.path" );

        path = path + ";.\\Compiler";

        System.setProperty( "java.class.path", path );
        System.out.println( "new java.class.path = " + System.getProperty( "java.class.path" ) );*/

        try
        {
            FileInputStream file = new FileInputStream( propertiesFile );
            properties.load( file );
            file.close();
        }
        catch( MalformedURLException e )
        {
            e.printStackTrace();  //To change body of catch statement use Options | File Templates.
        }
        catch( IOException e )
        {
            e.printStackTrace();  //To change body of catch statement use Options | File Templates.
        }

        String outDir = properties.getProperty( "outputDirectory" );

        if( outDir != null )
             outputDirectory = outDir;
        
        outDir = properties.getProperty( "outputInfoDirectory" );

        if( outDir != null )
            outputInfoDirectory = outDir;
        

        outDir = properties.getProperty( "outputOBJDirectory" );

        if( outDir != null )
             outputOBJDirectory = outDir;

        outDir = properties.getProperty( "outputEXEDirectory" );

        if( outDir != null )
             outputEXEDirectory = outDir;

        try 
        {
            UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
            // If you want the System L&F instead, comment out the above line and
            // uncomment the following:
            // UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (Exception exc) 
        {
            System.err.println("Error loading L&F: " + exc);
        }
    
        setBorder(BorderFactory.createEtchedBorder());
        setLayout(new BorderLayout());
    
        // create the embedded JTextComponent
        sourceEditor = createEditor(true);
        abstractSyntaxEditor = createEditor(false);
        iCTreeEditor = createEditor(false);
        cCTreeEditor = createEditor(false);
        iSEditor = createEditor(false);
        asmEditor = createEditor(false);
        errorEditor = createEditor(false);
        
        // Add this as a listener for undoable edits.
        sourceEditor.getDocument().addUndoableEditListener(undoHandler);
        sourceEditor.addKeyListener(new KeyListener()
        {

            public void keyTyped(KeyEvent e)
            {
                switch(e.getKeyChar())
                {
                    case 25: //Ctrl-y
                        redoAction.actionPerformed(null);
                        break;
                    case 26: //Ctrl-z
                        undoAction.actionPerformed(null);
                        break;
                    default:
                }
            }

            public void keyPressed(KeyEvent e)
            {
            }

            public void keyReleased(KeyEvent e)
            {
            }
        
        }
        );
        
        errorEditor.addMouseListener(new MouseListener()
        {

            public void mouseClicked(MouseEvent e)
            {
                if(e.getClickCount()==2)
                {
                    String selText=errorEditor.getSelectedText();
                    int[] pos=new int[2];
                    if(isSourcePos(selText,pos))
                    {
                        int line=1, 
                            off=1;
                        char[] app = sourceEditor.getText().toCharArray();
                        int i,
                            //newLineOffset=0,
                            lastNonChar=0;
                        
                        for(i=0;i<app.length;i++)
                        {    
                            if(app[i]=='\n')
                            {
                                off=1;
                                line++;
                                //newLineOffset=i;
                            }
                            else 
                                off++;
                            if(!Character.isLetterOrDigit(app[i]))
                                lastNonChar=i;
                              
                            if(pos[1]==off && pos[0]==line)
                            {
                                int j=i;
                                jTabbedPane1.setSelectedIndex(0);
                                sourceEditor.grabFocus();
                                if(lastNonChar==i)
                                {
                                    for(j=i+1;j<app.length;j++)
                                        if(!Character.isLetterOrDigit(app[j]))
                                            break;
                                }
                                
                                //sourceEditor.select(lastNonChar+1, j+1);
                                sourceEditor.setSelectionStart(lastNonChar+1);
                                sourceEditor.setSelectionEnd(j+1);
                                char[] errorText=errorEditor.getText().toCharArray();
                                int start=0,end=0;
                                int countNL=0;
                                for(int k=errorEditor.getCaretPosition();k<errorText.length;k++)
                                {
                                    if(countNL==0)
                                        start=k+1;
                                    if(errorText[k]=='\n')
                                        countNL++;
                                    if(countNL==2)
                                    {    
                                        end=k;
                                        break;
                                    }
                                }
                                if(end<start)
                                    end=errorText.length;
                                status.setText("Line: "+line+", Column: "+off+"; ERROR: "+errorEditor.getText().substring(start, end));
                            }
                        }
                    }
                }
            }
            
            public void mousePressed(MouseEvent e)
            {
                // TODO Auto-generated method stub
                
            }

            public void mouseReleased(MouseEvent e)
            {
                // TODO Auto-generated method stub
                
            }

            public void mouseEntered(MouseEvent e)
            {
                // TODO Auto-generated method stub
                
            }

            public void mouseExited(MouseEvent e)
            {
                // TODO Auto-generated method stub
                
            }
        
        });
    
        // install the command table
        commands = new Hashtable();
        Action[] actions = getActions();
        for (int i = 0; i < actions.length; i++) 
        {
            Action a = actions[i];
            //commands.put(a.getText(Action.NAME), a);
            commands.put(a.getValue(Action.NAME), a);
        }
        
        sourceEditorScroller = new JScrollPane();
        abstractSyntaxEditorScroller = new JScrollPane();
        iCTreeEditorScroller = new JScrollPane();
        cCTreeEditorScroller = new JScrollPane();
        iSEditorScroller = new JScrollPane();
        asmEditorScroller = new JScrollPane();
        errorEditorScroller = new JScrollPane();
        
        JViewport port = sourceEditorScroller.getViewport();
        port.add(sourceEditor);
        try 
        {
            String vpFlag = resources.getString("ViewportBackingStore");
            Boolean bs = new Boolean(vpFlag);
            //port.setBackingStoreEnabled(bs.booleanValue());
            if( bs.booleanValue() == true )
                 port.setScrollMode( JViewport.BACKINGSTORE_SCROLL_MODE );
            else
                 port.setScrollMode( JViewport.BLIT_SCROLL_MODE );
        } catch (MissingResourceException mre) 
        {
            // just use the viewport default
        }
        
        jTabbedPane1 = new JTabbedPane();
        jTabbedPane1.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);
        
        
        
        abstractSyntaxEditorScroller.setViewportView(abstractSyntaxEditor);
        iCTreeEditorScroller.setViewportView(iCTreeEditor);
        cCTreeEditorScroller.setViewportView(cCTreeEditor);
        iSEditorScroller.setViewportView(iSEditor);
        asmEditorScroller.setViewportView(asmEditor);
        errorEditorScroller.setViewportView(errorEditor);
        
        jTabbedPane1.addTab("Source File", sourceEditorScroller);
        
        sourceEditor.addCaretListener(new CaretListener()
            {
                public void caretUpdate(CaretEvent e)
                {
                    int line=1, 
                        off=1,
                        wa=e.getDot();
                    char[] app = sourceEditor.getText().toCharArray();
                    int i,
                        xtr=0;
                    
                    for(i=0;i<wa;i++)
                        if(app[i]=='\n')
                        {
                            off=1;
                            line++;
                            xtr++;
                        }
                        else 
                            off++;          
                    status.setText("Line: "+line+", Column: "+off);
                }
            }
        );
        
    
        menuItems = new Hashtable();
        menubar = createMenubar();
        add("North", menubar);
        JPanel panel = new JPanel();
        panel.setLayout(new BorderLayout());    
        panel.add("North",createToolbar());
        panel.add("Center", jTabbedPane1);
        add("Center", panel);
        add("South", createStatusbar());
        
    }

    public static void main(String[] args) 
    {
        try 
        {
            String vers = System.getProperty("java.version");
            if (vers.compareTo("1.1.2") < 0) 
            {
                System.out.println("!!!WARNING: Swing must be run with a " +
                                   "1.1.2 or higher version VM!!!");
            }
            JFrame frame = new JFrame();
            frame.setTitle(resources.getString("Title")+" - New file");
            frame.setBackground(Color.lightGray);
            frame.getContentPane().setLayout(new BorderLayout());
            frame.getContentPane().add("Center", new JOTTOEditor());
            frame.addWindowListener(new AppCloser());            
            frame.pack();
            frame.setSize(500, 600);
            frame.show();
        } 
        catch (Throwable t) 
        {
            System.out.println("uncaught exception: " + t);
            t.printStackTrace();
        }
    }

    /**
     * Fetch the list of actions supported by this
     * editor.  It is implemented to return the list
     * of actions supported by the embedded JTextComponent
     * augmented with the actions defined locally.
     */
    public Action[] getActions() 
    {
        return TextAction.augmentList(sourceEditor.getActions(), defaultActions);
    }

    /**
     * Create an editor to represent the given document.  
     */
    protected JTextComponent createEditor(boolean enableInput) 
    {
        JTextComponent c = new JTextArea();
        c.setEditable(enableInput);
        c.setDragEnabled(true);
        c.setFont(new Font("monospaced", Font.PLAIN, 12));
        return c;
    }

    /** 
     * Fetch the editor contained in this panel
     */
    protected JTextComponent getSourceEditor() 
    {
        return sourceEditor;
    }

    /**
     * To shutdown when run as an application.  This is a
     * fairly lame implementation.   A more self-respecting
     * implementation would at least check to see if a save
     * was needed.
     */
    protected static final class AppCloser extends WindowAdapter 
    {
        public void windowClosing(WindowEvent e) 
        {
            System.exit(0);
        }
    }

    /**
     * Find the hosting frame, for the file-chooser dialog.
     */
    protected Frame getFrame() 
    {
        for (Container p = getParent(); p != null; p = p.getParent()) 
        {
            if (p instanceof Frame) 
            {
                return (Frame) p;
            }
        }
        return null;
    }

    /**
     * This is the hook through which all menu items are
     * created.  It registers the result with the menuitem
     * hashtable so that it can be fetched with getMenuItem().
     * @see #getMenuItem
     */
    protected JMenuItem createMenuItem(String cmd) 
    {
        JMenuItem mi = new JMenuItem(getResourceString(cmd + labelSuffix));
        URL url = getResource(cmd + imageSuffix);
        if (url != null) 
        {
            mi.setHorizontalTextPosition(JButton.RIGHT);
            mi.setIcon(new ImageIcon(url));
        }
        String astr = getResourceString(cmd + actionSuffix);
        if (astr == null) 
            astr = cmd;
        mi.setActionCommand(astr);
        Action a = getAction(astr);
        if (a != null) 
        {
            mi.addActionListener(a);
            a.addPropertyChangeListener(createActionChangeListener(mi));
            mi.setEnabled(a.isEnabled());
        } 
        else 
        {
            mi.setEnabled(false);
        }
        menuItems.put(cmd, mi);
        return mi;
    }

    /**
     * Fetch the menu item that was created for the given
     * command.
     * @param cmd  Name of the action.
     * @returns item created for the given command or null
     *  if one wasn't created.
     */
    protected JMenuItem getMenuItem(String cmd) 
    {
        return (JMenuItem) menuItems.get(cmd);
    }

    protected Action getAction(String cmd) 
    {
        return (Action) commands.get(cmd);
    }

    protected String getResourceString(String nm) 
    {
        String str;
        try 
        {
            str = resources.getString(nm);
        } 
        catch (MissingResourceException mre) 
        {
            str = null;
        }
        return str;
    }

    protected URL getResource(String key) 
    {
        String name = getResourceString(key);
        if (name != null) 
        {
            URL url = this.getClass().getResource(name);
            return url;
        }
        return null;
    }

    protected Container getToolbar() 
    {
        return toolbar;
    }

    protected JMenuBar getMenubar() 
    {
        return menubar;
    }

    /**
     * Create a status bar
     */
    protected Component createStatusbar() 
    {
        // need to do something reasonable here
        status = new StatusBar();
        return status;
    }

    /**
     * Resets the undo manager.
     */
    protected void resetUndoManager() 
    {
        undo.discardAllEdits();
        undoAction.update();
        redoAction.update();
    }

    /**
     * Create the toolbar.  By default this reads the 
     * resource file for the definition of the toolbar.
     */
    private Component createToolbar() 
    {
        toolbar = new JToolBar();
        String[] toolKeys = tokenize(getResourceString("toolbar"));
        for (int i = 0; i < toolKeys.length; i++) 
        {
            if (toolKeys[i].equals("-")) 
                toolbar.add(Box.createHorizontalStrut(5));
            else
                toolbar.add(createTool(toolKeys[i]));
        }
        toolbar.add(Box.createHorizontalGlue());
        return toolbar;
    }

    /**
     * Hook through which every toolbar item is created.
     */
    protected Component createTool(String key) 
    {
        return createToolbarButton(key);
    }

    /**
     * Create a button to go inside of the toolbar.  By default this
     * will load an image resource.  The image filename is relative to
     * the classpath (including the '.' directory if its a part of the
     * classpath), and may either be in a JAR file or a separate file.
     * 
     * @param key The key in the resource file to serve as the basis
     *  of lookups.
     */
    protected JButton createToolbarButton(String key) 
    {
        URL url = getResource(key + imageSuffix);
        JButton b = new JButton(new ImageIcon(url)) 
                    {
                        public float getAlignmentY() { return 0.5f; }
                    };
        b.setRequestFocusEnabled(false);
        b.setMargin(new Insets(1,1,1,1));
    
        String astr = getResourceString(key + actionSuffix);
        if (astr == null)
            astr = key;
        
        Action a = getAction(astr);
        if (a != null) 
        {
            b.setActionCommand(astr);
            b.addActionListener(a);
        } 
        else 
            b.setEnabled(false);
        
    
        String tip = getResourceString(key + tipSuffix);
        if (tip != null) 
            b.setToolTipText(tip);
     
        return b;
    }

    /**
     * Take the given string and chop it up into a series
     * of strings on whitespace boundries.  This is useful
     * for trying to get an array of strings out of the
     * resource file.
     */
    protected String[] tokenize(String input) 
    {
        Vector v = new Vector();
        StringTokenizer t = new StringTokenizer(input);
        String cmd[];
    
        while (t.hasMoreTokens())
            v.addElement(t.nextToken());
        cmd = new String[v.size()];
        for (int i = 0; i < cmd.length; i++)
            cmd[i] = (String) v.elementAt(i);
    
        return cmd;
    }

    /**
     * Create the menubar for the app.  By default this pulls the
     * definition of the menu from the associated resource file. 
     */
    protected JMenuBar createMenubar() 
    {
        //JMenuItem mi;
        JMenuBar mb = new JMenuBar();
    
        String[] menuKeys = tokenize(getResourceString("menubar"));
        for (int i = 0; i < menuKeys.length; i++) 
        {
            JMenu m = createMenu(menuKeys[i]);
            if (m != null)
                mb.add(m);
        }
        return mb;
    }

    /**
     * Create a menu for the app.  By default this pulls the
     * definition of the menu from the associated resource file.
     */
    protected JMenu createMenu(String key) 
    {
        String[] itemKeys = tokenize(getResourceString(key));
        JMenu menu = new JMenu(getResourceString(key + "Label"));
        for (int i = 0; i < itemKeys.length; i++) 
        {
            if (itemKeys[i].equals("-"))
                menu.addSeparator();
            else 
            {
                JMenuItem mi = createMenuItem(itemKeys[i]);
                menu.add(mi);
            }
        }
        return menu;
    }

    /*
        utilities
    */
    public String deleteExtension( String fileName )
    {
        char[] file = fileName.toCharArray();

        for( int i = fileName.length() - 1; i>=0; i-- )
           if( file[i] == '.' )
                return fileName.substring( 0, i );

        return fileName;
    }

    // Yarked from JMenu, ideally this would be public.
    protected PropertyChangeListener createActionChangeListener(JMenuItem b) 
    {
        return new ActionChangedListener(b);
    }
    
    /**
     * 
     */
    private void showCompileInfos()
    {
        
        String outName = deleteExtension( currentDocumentFileName );
        String[] segNames=
        {
            "Results",
            "Abstract Syntax",
            "Intermediate Code Tree",
            "Canonical Code Tree",
            "Instruction Selection",
        };
        
        JTextComponent[] editors=
        {
            errorEditor,
            abstractSyntaxEditor,
            iCTreeEditor,
            cCTreeEditor,
            iSEditor,
        };
        loadInfosFromFile(outputInfoDirectory+outName+".txt", "---- ", segNames, editors);

        if(!errorEditor.getText().equals(""))
            jTabbedPane1.addTab("Build Results", errorEditorScroller);
        if(!abstractSyntaxEditor.getText().equals(""))
            jTabbedPane1.addTab("Abstract Syntax Tree", abstractSyntaxEditorScroller);
        if(!iCTreeEditor.getText().equals(""))
            jTabbedPane1.addTab("Intermediate Code Tree", iCTreeEditorScroller);
        if(!cCTreeEditor.getText().equals(""))
            jTabbedPane1.addTab("Canonical Code Tree", cCTreeEditorScroller);
        if(!iSEditor.getText().equals(""))
            jTabbedPane1.addTab("Instruction Selection", iSEditorScroller);
        if(compileSuccessful)
        {    
            jTabbedPane1.addTab("Assembly Code", asmEditorScroller);
            loadInfosFromFile(outputDirectory+outName+".asm", asmEditor);
        }
        jTabbedPane1.setSelectedIndex(1);
    }

    /**
     * @param fileName
     * @param separator
     * @param segNames
     * @param editors
     * @param sorted
     */
    private void loadInfosFromFile(String fileName, String separator, String[] segNames, JTextComponent[] editors)
    {
        try
        {
            BufferedReader br=new BufferedReader(new FileReader(fileName));
            StringBuffer text=new StringBuffer(100000);
            int segCount=0;
            while(br.ready())
            {
                String line=br.readLine();
                if(line.startsWith(separator) && line.substring(separator.length()).equals(segNames[segCount]))
                    break;
            }
            boolean go=br.ready();
            while(go)
            {
                String line=br.readLine();
                
                if(line.startsWith(separator) ||
                   !br.ready())
                {  
                    editors[segCount].setText(text.toString());
                    editors[segCount].setCaretPosition(0);
                    segCount++;
                    if(segCount==segNames.length || !br.ready())
                        break;
                    while(!line.startsWith(separator) || !line.substring(separator.length()).equals(segNames[segCount]))
                    {    
                        line=br.readLine();
                        if(!br.ready())
                            return;
                    }
                    text.delete(0,text.length());
                    continue;
                }
                if(text.length()>0)
                    text.append("\n");
                text.append(line);
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
    }

    /**
     * @param fileName
     * @param editor
     */
    private void loadInfosFromFile(String fileName, JTextComponent editor)
    {
        try
        {
            BufferedReader br=new BufferedReader(new FileReader(fileName));
            String text="";
            
            while(br.ready())
            {
                String line=br.readLine();
                if(text.length()>0)
                    text+="\n";
                text+=line;
                
            }
            editor.setText(text);
            editor.setCaretPosition(0);
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
    }

    /**
     * @param fileName
     * @param separator
     * @param segmentName
     * @param editor
     */
    private void loadInfosFromFile(String fileName, String separator, String segmentName, JTextComponent editor)
    {
        try
        {
            BufferedReader br=new BufferedReader(new FileReader(fileName));
            String text="";
            boolean stilldoing=br.ready();
            while(stilldoing)
            {
                String line=br.readLine();
                if(line.startsWith(separator))
                {
                    if(line.substring(separator.length()).equals(segmentName))
                    {
                        while(br.ready())
                        {
                            line=br.readLine();
                            if(line.startsWith(separator))
                            {
                                stilldoing=false;
                                break;
                            }
                            if(text.length()>0)
                                text+="\n";
                            text+=line;
                        }
                    }
                }
                stilldoing&=br.ready();
            }
            editor.setText(text);
            editor.setCaretPosition(0);
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
    }

    /**
     * 
     */
    private void resetEditors()
    {
        sourceEditor.setDocument(new PlainDocument());
        
        abstractSyntaxEditor.setText("");
        iCTreeEditor.setText("");
        cCTreeEditor.setText("");
        iSEditor.setText("");
        asmEditor.setText("");
        errorEditor.setText("");
        
        if(jTabbedPane1.getTabCount()<2)
            return;
        jTabbedPane1.removeTabAt(jTabbedPane1.indexOfTab("Build Results"));
        if(jTabbedPane1.getTabCount()<2)
            return;        
        jTabbedPane1.removeTabAt(jTabbedPane1.indexOfTab("Abstract Syntax Tree"));
        if(jTabbedPane1.getTabCount()<2)
            return;        
        jTabbedPane1.removeTabAt(jTabbedPane1.indexOfTab("Intermediate Code Tree"));
        jTabbedPane1.removeTabAt(jTabbedPane1.indexOfTab("Canonical Code Tree"));
        jTabbedPane1.removeTabAt(jTabbedPane1.indexOfTab("Instruction Selection"));
        jTabbedPane1.removeTabAt(jTabbedPane1.indexOfTab("Assembly Code"));        
    }
        
        
    // Yarked from JMenu, ideally this would be public.
    private class ActionChangedListener implements PropertyChangeListener 
    {
        JMenuItem menuItem;

        ActionChangedListener(JMenuItem mi) 
        {
            super();
            this.menuItem = mi;
        }
        
        public void propertyChange(PropertyChangeEvent e) 
        {
            String propertyName = e.getPropertyName();
            if (e.getPropertyName().equals(Action.NAME)) 
            {
                String text = (String) e.getNewValue();
                menuItem.setText(text);
            } 
            else 
                if (propertyName.equals("enabled")) 
                {
                    Boolean enabledState = (Boolean) e.getNewValue();
                    menuItem.setEnabled(enabledState.booleanValue());
                }
        }
    }

    private JTextComponent sourceEditor;
    private JTextComponent abstractSyntaxEditor;
    private JTextComponent iCTreeEditor;
    private JTextComponent cCTreeEditor;
    private JTextComponent iSEditor;
    private JTextComponent asmEditor;
    private JTextComponent errorEditor;
    
    
    JScrollPane sourceEditorScroller;
    JScrollPane abstractSyntaxEditorScroller;
    JScrollPane iCTreeEditorScroller;
    JScrollPane cCTreeEditorScroller;
    JScrollPane iSEditorScroller;
    JScrollPane asmEditorScroller;
    JScrollPane errorEditorScroller;
    
    private JTabbedPane jTabbedPane1;
    private Hashtable commands;
    private Hashtable menuItems;
    private JMenuBar menubar;
    private JToolBar toolbar;
    private StatusBar status;
    //private JFrame elementTreeFrame;
    protected ElementTreePanel elementTreePanel;

    protected FileDialog fileDialog;
    protected SettingsDialog settingsDialog;

    /**
     * Listener for the edits on the current document.
     */
    protected UndoableEditListener undoHandler = new UndoHandler();

    /** UndoManager that we add edits to. */
    protected UndoManager undo = new UndoManager();

    /**
     * Suffix applied to the key used in resource file
     * lookups for an image.
     */
    public static final String imageSuffix = "Image";

    /**
     * Suffix applied to the key used in resource file
     * lookups for a label.
     */
    public static final String labelSuffix = "Label";

    /**
     * Suffix applied to the key used in resource file
     * lookups for an action.
     */
    public static final String actionSuffix = "Action";

    /**
     * Suffix applied to the key used in resource file
     * lookups for tooltip text.
     */
    public static final String tipSuffix = "Tooltip";

    public static final String openAction = "open";
    public static final String newAction  = "new";
    public static final String saveAsAction = "saveAs";
    public static final String saveAction = "save";    
    public static final String exitAction = "exit";
    public static final String showElementTreeAction = "showElementTree";
    public static final String compileAction = "compile";
    public static final String runAction = "run";
    public static final String settingsAction = "settings";
    public static final String linkAction = "link";
    public static final String infoAction = "info";
    
    
    class UndoHandler implements UndoableEditListener 
    {

    /**
     * Messaged when the Document has created an edit, the edit is
     * added to <code>undo</code>, an instance of UndoManager.
     */
        public void undoableEditHappened(UndoableEditEvent e) 
        {
            undo.addEdit(e.getEdit());
            undoAction.update();
            redoAction.update();
        }
    }

    
    class StatusBar extends JPanel 
    {
        JTextArea text;
        public StatusBar() 
        {
            super();
            setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
            text=new JTextArea();
            text.setBackground(Color.lightGray);
            add(text);
            text.setEditable(false);
        }
        
        public String getText()
        {
            return text.getText();
        }
        
        public void setText(String txt)
        {
            text.setText(txt);
        }

        public void paint(Graphics g) 
        {
            super.paint(g);
        }
    }

    // --- action implementations -----------------------------------

    private UndoAction undoAction = new UndoAction();
    private RedoAction redoAction = new RedoAction();
    
    Action saveAsAct;

    /**
     * Actions defined by the JOTTOEditor class
     */
    private Action[] defaultActions = 
    {
        new NewAction(),
        saveAsAct=new SaveAsAction(),
        new SaveAction(saveAsAct),
        new OpenAction(),
        new ExitAction(),
        new CompileAction(),
        new SettingsAction(),
        new LinkAction(),
        new RunAction(),
        new InfoAction(),
            undoAction,
            redoAction
    };

    class UndoAction extends AbstractAction 
    {
        public UndoAction() 
        {
            super("Undo");
            setEnabled(false);
        }
    
        public void actionPerformed(ActionEvent e) 
        {
            try 
            {
                undo.undo();
            } 
            catch (CannotUndoException ex) 
            {
                System.out.println("Unable to undo: " + ex);
                //ex.printStackTrace();
            }
            update();
            redoAction.update();
        }
    
        protected void update() 
        {
            if(undo.canUndo()) 
            {
                setEnabled(true);
                putValue(Action.NAME, undo.getUndoPresentationName());
            }
            else 
            {
                setEnabled(false);
                putValue(Action.NAME, "Undo");
            }
        }
    }

    class RedoAction extends AbstractAction 
    {
        public RedoAction() 
        {
            super("Redo");
            setEnabled(false);
        }
    
        public void actionPerformed(ActionEvent e) 
        {
            try 
            {
                undo.redo();
            } 
            catch (CannotRedoException ex) 
            {
                System.out.println("Unable to redo: " + ex);
                //ex.printStackTrace();
            }
            update();
            undoAction.update();
        }
    
        protected void update() 
        {
            if(undo.canRedo()) 
            {
                setEnabled(true);
                putValue(Action.NAME, undo.getRedoPresentationName());
            }
            else 
            {
                setEnabled(false);
                putValue(Action.NAME, "Redo");
            }
        }
    }

    class OpenAction extends NewAction 
    {

        OpenAction() 
        {
            super(openAction);
        }

        public void actionPerformed(ActionEvent e) 
        {
            Frame frame = getFrame();
            if (fileDialog == null) 
                fileDialog = new FileDialog(frame);
        
            String sourceDir = properties.getProperty( "sourceDirectory" );
            if( sourceDir != null )
                 fileDialog.setDirectory( sourceDir );

            fileDialog.setMode(FileDialog.LOAD);
            fileDialog.setTitle("Load file");
            fileDialog.show();
    
            String file = fileDialog.getFile();
            if (file == null)
                return;
            
            String directory = fileDialog.getDirectory();
    
            properties.setProperty( "sourceDirectory", directory );
    
            currentDocumentFileName = file;
            currentDocumentFileDirectory = directory;
    
            File f = new File(directory, file);
            if (f.exists()) 
            {
                Document oldDoc = getSourceEditor().getDocument();
                if(oldDoc != null)
                    oldDoc.removeUndoableEditListener(undoHandler);
                if (elementTreePanel != null)
                    elementTreePanel.setEditor(null);

                //getSourceEditor().setDocument(new PlainDocument());
                resetEditors();
                frame.setTitle(resources.getString("Title")+" - "+file);
                Thread loader = new FileLoader(f, sourceEditor.getDocument());
                loader.start();
            }
        }
    }

    /*
       save action
    */
    class SaveAction extends NewAction 
    {

        static final int BUFFER_DIM = 4096;
        
        private Action saveAs;

        SaveAction(Action save_as) 
        {
            super(saveAction);
            saveAs=save_as;
        }

        public String convertReturns( String source )
        {
            /*if( source.indexOf( "\r\n" ) != -1 )
                 return source;

            String[] phrases = source.split( "\n" );
            String result = "";

            for( int i=0; i<phrases.length; i++ )
               result = result + phrases[i] + "\r\n";

            return result;*/

            String result = "";

            for( int i=0; i<source.length(); i++ )
            {
                char c = source.charAt( i );

                if( i == 0 )
                {
                    if( c == '\r' && source.charAt( i+1 ) != '\n' )
                         result += "\r\n";
                    else
                         if( c == '\n' )
                              result += "\r\n";
                         else
                              result += c;
                }
                else // se i > 0
                     if( i == source.length() - 1 )
                     {
                         if( c == '\r' )
                              result += "\r\n";
                         else
                              if( c == '\n' && source.charAt( i-1 ) != '\r')
                                   result += "\r\n";
                              else
                                   result += c;
                     }
                     else // 0 < i < lunghezzaStringa
                          if( c == '\r' && source.charAt( i+1 ) != '\n' )
                               result += "\r\n";
                          else
                               if( c == '\n' && source.charAt( i-1 ) != '\r')
                                    result += "\r\n";
                               else
                                    result += c;
            }

            return result;
        }

        public void actionPerformed(ActionEvent e) 
        {
            if(currentDocumentFileName==null)
            {    
                saveAs.actionPerformed(e);
                return;
            }
            
            File f = new File(currentDocumentFileDirectory, currentDocumentFileName );
            FileWriter outF = null;

            try 
            {
                outF = new FileWriter( f );
            } 
            catch (IOException e1) 
            {
                e1.printStackTrace();  //To change body of catch statement use Options | File Templates.
            }

            if( outF == null )
                 return;

            Document doc = getSourceEditor().getDocument();
            int pos = 0;

            try 
            {
                while( pos < doc.getLength() )
                {
                    int byteToWrite;
                    if( pos + BUFFER_DIM > doc.getLength() )
                         byteToWrite = doc.getLength() - pos;
                    else
                         byteToWrite = BUFFER_DIM;

                    String buffer = doc.getText( pos, byteToWrite );
                    outF.write( convertReturns( buffer ) );
                    pos += buffer.length();
                }
                outF.close();
            } 
            catch (BadLocationException e1) 
            {
                e1.printStackTrace();  //To change body of catch statement use Options | File Templates.
            } catch (IOException e1) 
            {
                e1.printStackTrace();  //To change body of catch statement use Options | File Templates.
            }
        }
    }
    
    /*
     save as action
     */
    class SaveAsAction extends NewAction 
    {

        static final int BUFFER_DIM = 4096;

        SaveAsAction() 
        {
            super(saveAsAction);
        }

        public String convertReturns( String source )
        {
            /*if( source.indexOf( "\r\n" ) != -1 )
             return source;

             String[] phrases = source.split( "\n" );
             String result = "";

             for( int i=0; i<phrases.length; i++ )
             result = result + phrases[i] + "\r\n";

             return result;*/

            String result = "";

            for( int i=0; i<source.length(); i++ )
               {
                char c = source.charAt( i );

                if( i == 0 )
                   {
                    if( c == '\r' && source.charAt( i+1 ) != '\n' )
                        result += "\r\n";
                    else
                        if( c == '\n' )
                            result += "\r\n";
                        else
                            result += c;
                }
                else // se i > 0
                    if( i == source.length() - 1 )
                       {
                        if( c == '\r' )
                            result += "\r\n";
                        else
                            if( c == '\n' && source.charAt( i-1 ) != '\r')
                                result += "\r\n";
                            else
                                result += c;
                    }
                    else // 0 < i < lunghezzaStringa
                        if( c == '\r' && source.charAt( i+1 ) != '\n' )
                            result += "\r\n";
                        else
                            if( c == '\n' && source.charAt( i-1 ) != '\r')
                                result += "\r\n";
                            else
                                result += c;
            }

            return result;
        }

        public void actionPerformed(ActionEvent e) 
        {
            Frame frame = getFrame();
            if (fileDialog == null)
                fileDialog = new FileDialog(frame);
            

            String sourceDir = properties.getProperty("sourceDirectory");
            if (sourceDir != null)
                fileDialog.setDirectory(sourceDir);

            fileDialog.setMode(FileDialog.SAVE);
            fileDialog.setTitle("Save file");
            fileDialog.show();

            String file = fileDialog.getFile();
            if (file == null)
                return;
            
            String directory = fileDialog.getDirectory();

            properties.setProperty("sourceDirectory", directory);

            currentDocumentFileName = file;
            currentDocumentFileDirectory = directory;

            File f = new File(directory, file);
            FileWriter outF = null;

            try 
            {
                outF = new FileWriter( f );
            } 
            catch (IOException e1) 
            {
                e1.printStackTrace();  //To change body of catch statement use Options | File Templates.
            }

            if( outF == null )
                return;

            Document doc = getSourceEditor().getDocument();
            int pos = 0;

            try 
            {
                while( pos < doc.getLength() )
                   {
                    int byteToWrite;
                    if( pos + BUFFER_DIM > doc.getLength() )
                        byteToWrite = doc.getLength() - pos;
                    else
                        byteToWrite = BUFFER_DIM;

                    String buffer = doc.getText( pos, byteToWrite );
                    outF.write( convertReturns( buffer ) );
                    pos += buffer.length();
                }
                outF.close();
            } 
            catch (BadLocationException e1) 
            {
                e1.printStackTrace();  //To change body of catch statement use Options | File Templates.
            } catch (IOException e1) 
            {
                e1.printStackTrace();  //To change body of catch statement use Options | File Templates.
            }

            frame.setTitle(resources.getString("Title")+" - "+file);
        }
    }
    

    class NewAction extends AbstractAction 
    {
    
        NewAction() 
        {
            super(newAction);
        }
    
        NewAction(String nm) 
        {
            super(nm);
        }
    
        public void actionPerformed(ActionEvent e) 
        {
            Document oldDoc = getSourceEditor().getDocument();
            if(oldDoc != null)
                oldDoc.removeUndoableEditListener(undoHandler);

            currentDocumentFileName = null;
            currentDocumentFileDirectory = null;

            resetEditors();
            sourceEditor.getDocument().addUndoableEditListener(undoHandler);
            resetUndoManager();
            revalidate();
            Frame f=getFrame();
            f.setTitle(resources.getString("Title")+" - New file");
        }        
    }

    /**
     * Really lame implementation of an exit command
     */
    class ExitAction extends AbstractAction 
    {

        ExitAction() 
        {
            super(exitAction);
        }

        public void actionPerformed(ActionEvent e) 
        {
            try
            {
                FileOutputStream file = new FileOutputStream( propertiesFile );
                properties.store( file, "settings configure" );
                file.close();
            }
            catch( FileNotFoundException e1 ) {}
            catch( IOException e1 )
            {
                e1.printStackTrace();  //To change body of catch statement use Options | File Templates.
            }
            System.exit(0);
        }
    }
    
    static boolean compileSuccessful=false;
    static boolean linkSuccessful=false;

    /*
       compile action
    */
    class CompileAction extends AbstractAction
    {
        Class compilerClass;

        CompileAction()
        {
            super( compileAction );

            try 
            {
                compilerClass = ClassLoader.getSystemClassLoader().loadClass( "Main.Jotto" );
            } 
            catch (ClassNotFoundException e) 
            {
                e.printStackTrace();  //To change body of catch statement use Options | File Templates.
            }
        }


        public void actionPerformed( ActionEvent e )
        {
            /*try
            {*/
                /*PrintStream syserr=, sysout;
                System.setOut(out);
                System.setErr(err);*/
                String[] argv = new String[8];

                if( currentDocumentFileName == null && currentDocumentFileDirectory == null)
                {    
                    saveAsAct.actionPerformed(e);
                    if(currentDocumentFileName == null)
                        return;
                }
                
                argv[0]="-a";
                argv[1]="-i";
                argv[2]="-c";
                argv[3]="-n";
                argv[4]="-e";
                argv[5] = "\""+currentDocumentFileDirectory + currentDocumentFileName+"\"";
                argv[6] = "\""+outputDirectory + deleteExtension( currentDocumentFileName ) + ".asm\"";
                argv[7] = "\""+outputInfoDirectory + deleteExtension( currentDocumentFileName ) + ".txt\"";


                File asm = new File( argv[1] );
                if( asm.exists() )
                    asm.delete();
                

                if( compilerClass != null )
                {
                    //Parse.Main compiler = (Parse.Main)(compilerClass.newInstance());
                    //if( compiler != null )
                    
                    /*Jotto.setErrStream(outerr);
                    Jotto.setOutStream(outerr);*/

                    //compiler.classMainMethod( argv );
                    String cmd="jotto.bat ";
                    for(int i=0;i<argv.length;i++)
                        cmd+=argv[i]+" ";
                    int errors=1;
                    
                    try
                    {
                        Process proc=Runtime.getRuntime().exec( cmd );
                        proc.waitFor();
                        errors= proc.exitValue();
                    }
                    catch (Exception e1)
                    {
                        e1.printStackTrace();
                    }
                    compileSuccessful= errors==0 ;                    
                }
                showCompileInfos();
/*
            }
            catch( InstantiationException e1 )
            {
                e1.printStackTrace();  //To change body of catch statement use Options | File Templates.
            }
            catch( IllegalAccessException e1 )
            {
                e1.printStackTrace();  //To change body of catch statement use Options | File Templates.
            }
            catch( IOException e1 )
            {
                e1.printStackTrace();  //To change body of catch statement use Options | File Templates.
            }*/
        }        
    }

    /*
       link action
    */
    class LinkAction extends AbstractAction
    {
        LinkAction()
        {
            super( linkAction );
        }

        public void actionPerformed( ActionEvent e )
        {
            if( outputDirectory == null || outputInfoDirectory == null || outputOBJDirectory == null || outputEXEDirectory == null ||currentDocumentFileName == null )
                return;

            Action compile= getAction( "compile" );
            if( compile != null )
            {    
                 compile.actionPerformed( e );
                 if(!compileSuccessful)
                 {
                     return;  
                 }
            }

            String outFile = outputDirectory + deleteExtension( currentDocumentFileName );
            String infoFile = outputInfoDirectory + deleteExtension( currentDocumentFileName );
            String objFile = outputOBJDirectory + deleteExtension( currentDocumentFileName );
            String exeFile = outputEXEDirectory + deleteExtension( currentDocumentFileName );

            if( new File( outFile + ".asm" ).exists() != true )
                return;

            try 
            {
                String cmd = ".\\bin\\asm2exe4editor.bat \"" + outFile + ".asm\" \"" + objFile + ".obj\" \"" + exeFile + ".exe\">\""+infoFile+".compiledebug.txt\""; //>\""+infoFile+".compiledebug.txt\"
                Process p = Runtime.getRuntime().exec( cmd );
                p.waitFor();
                if(!(linkSuccessful = (p.exitValue() == 0)))
                {
                    JOptionPane.showConfirmDialog( null, "LINK NON ESEGUITO", "COMPILAZIONE", JOptionPane.DEFAULT_OPTION, JOptionPane.ERROR_MESSAGE);
                    return;
                }
                /*String cmd1 = ".\\bin\\lcc \""+outFile+".asm\" -Fo\""+objFile+".obj\" -g5 -A >NULL";
                String cmd2 = ".\\bin\\lcclnk \""+objFile+".obj\" .\\bin\\runtime.obj -o \""+exeFile+".exe\">>NULL";
                Process p1 = Runtime.getRuntime().exec( cmd1 );                
                p1.waitFor();
                
                if(!(linkSuccessful = (p1.exitValue() == 0)))
                {
                    JOptionPane.showConfirmDialog( null, "LINK NON ESEGUITO", "COMPILAZIONE", JOptionPane.DEFAULT_OPTION, JOptionPane.ERROR_MESSAGE);
                    return;
                }
                Process p2 = Runtime.getRuntime().exec( cmd2 );
                p2.waitFor();
                
                if(!(linkSuccessful = (p2.exitValue() == 0)))
                {
                    JOptionPane.showConfirmDialog( null, "LINK NON ESEGUITO", "COMPILAZIONE", JOptionPane.DEFAULT_OPTION, JOptionPane.ERROR_MESSAGE);
                    return;
                }*/
            } 
            catch (Exception e1) 
            {
                e1.printStackTrace();  //To change body of catch statement use Options | File Templates.
            }
        }

    }

    /*
       run action
    */
    class RunAction extends AbstractAction
    {
        RunAction()
        {
            super( runAction );
        }

        public void actionPerformed( ActionEvent e )
        {
            if( outputEXEDirectory == null ||currentDocumentFileName == null )
                return;

            Action link = getAction( "link" );
            if( link != null )
            {    
                link.actionPerformed( e );
                if(!compileSuccessful || !linkSuccessful)
                    return;
            }

            String exeFile = outputEXEDirectory + deleteExtension( currentDocumentFileName );
            String outFile = outputDirectory + deleteExtension( currentDocumentFileName );

            if( new File( outFile + ".asm" ).exists() != true )
                return;
            
            try 
            {
                String cmd = ".\\bin\\execul \""+exeFile+".exe\"";
                Runtime.getRuntime().exec( cmd );
            } 
            catch (IOException e1) 
            {
                e1.printStackTrace();  //To change body of catch statement use Options | File Templates.
            }
          
        }

    }

    /*
       settings action
    */
    class SettingsAction extends AbstractAction
    {
        SettingsAction()
        {
            super( settingsAction );
        }

        public void actionPerformed( ActionEvent e )
        {
            Frame frame = getFrame();
            if( settingsDialog == null )
            {
                settingsDialog = new SettingsDialog( frame );
            }
            settingsDialog.show();
        }

    }

    /*
       help action
    */
    class InfoAction extends AbstractAction
    {
        InfoAction()
        {
            super( infoAction );
        }

        public void actionPerformed( ActionEvent e )
        {
            Frame frame = getFrame();

            JOptionPane.showMessageDialog( frame, "JOTTOEditor 1.0\n        for\n JOTTO (Java Oils and Touches a Tiger cOmpiler)\n        by\nCristiano Cenci\nGiovanni Ottieri\nCristiano Tagliamonte", "INFO", JOptionPane.INFORMATION_MESSAGE );
        }

    }

    /**
     * Thread to load a file into the text storage model
     */
    class FileLoader extends Thread 
    {

        FileLoader(File f, Document doc) 
        {
            setPriority(4);
            this.f = f;
            this.doc = doc;
        }

        public void run() 
        {
            try 
            {
                // initialize the statusbar
                /*status.removeAll();
                JProgressBar progress = new JProgressBar();
                progress.setMinimum(0);
                progress.setMaximum((int) f.length());
                status.add(progress);
                status.revalidate();*/
        
                // try to start reading
                Reader in = new FileReader(f);
                char[] buff = new char[4096];
                int nch;
                String curstr;
                while ((nch = in.read(buff, 0, buff.length)) != -1) 
                {
                    curstr=new String(buff, 0, nch);
                    doc.insertString(doc.getLength(), removeNewLines(curstr), null);
                    //progress.setValue(progress.getValue() + nch);
                }
        
                // we are done... get rid of progressbar
                doc.addUndoableEditListener(undoHandler);
                /*status.removeAll();
                status.revalidate();*/
        
                resetUndoManager();
            }
            catch (IOException e) 
            {
                System.err.println(e.toString());
            }
            catch (BadLocationException e) 
            {
                System.err.println(e.getMessage());
            }
            if (elementTreePanel != null) 
            {
                SwingUtilities.invokeLater(new Runnable() 
                    {
                        public void run() 
                        {
                            elementTreePanel.setEditor(getSourceEditor());
                        }
                    });
            }
        }

        /**
         * @param curstr
         * @return
         */
        private String removeNewLines(String curstr)
        {
            for(int i=0;i<curstr.length();i++)
            {
                if(curstr.charAt(i)=='\r')
                    curstr=curstr.substring(0,i)+curstr.substring(i--+1);
            }
            return curstr;
        }

        Document doc;
        File f;
    }

    public class SettingsDialog extends JDialog
    {
        final JTextField text    = new JTextField( 20 );
        final JTextField textInfo    = new JTextField( 20 );
        final JTextField textOBJ = new JTextField( 20 );
        final JTextField textEXE = new JTextField( 20 );

        public SettingsDialog( Frame frame )
        {
            super( frame );

            String outputDir = properties.getProperty( "outputDirectory" );
            if( outputDir != null )
            {
                outputDirectory = outputDir;
                text.setText( outputDirectory );
            }
            
            outputDir = properties.getProperty( "outputInfoDirectory" );
            if( outputDir != null )
            {
                outputInfoDirectory = outputDir;
                textInfo.setText( outputInfoDirectory );
            }

            outputDir = properties.getProperty( "outputOBJDirectory" );
            if( outputDir != null )
            {
                outputOBJDirectory = outputDir;
                textOBJ.setText( outputOBJDirectory );
            }

            outputDir = properties.getProperty( "outputEXEDirectory" );
            if( outputDir != null )
            {
                outputEXEDirectory = outputDir;
                textEXE.setText( outputEXEDirectory );
            }

            JPanel outputDestPanel;
            JPanel outputInfoPanel;
            JPanel outputOBJPanel;
            JPanel outputEXEPanel;
            JPanel buttonPanel;

            JButton browseButton = new JButton( "browse..." );
            browseButton.addActionListener( new AbstractAction() 
                {
                    public void actionPerformed( ActionEvent e )
                    {
                       /* FileDialog fileDialog = new FileDialog( getFrame() );
    
                        fileDialog.setDirectory( outputDirectory );
                        fileDialog.setMode( FileDialog.LOAD );
                        fileDialog.setTitle("Load file");
                        fileDialog.show();
    
                        String directory = fileDialog.getDirectory();
                        text.setText( directory );*/
                        JFileChooser chooser = new JFileChooser();
                        chooser.setCurrentDirectory(new File(text.getText()));
                        chooser.setDialogType(JFileChooser.OPEN_DIALOG);
                        chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
                        
                        
                        int returnVal = chooser.showOpenDialog(getFrame());
                        if(returnVal == JFileChooser.APPROVE_OPTION) 
                        {
                            String directory = chooser.getSelectedFile().getAbsolutePath();
                            text.setText( directory+"\\" );
                        }                               
                    }
                }
            );
            getContentPane().add( outputDestPanel = new JPanel() );
            outputDestPanel.add( new JLabel( "output ASM directory" ) );
            outputDestPanel.add( text );
            outputDestPanel.add( browseButton );

            JButton browseButtonInfo = new JButton( "browse..." );
            browseButtonInfo.addActionListener( new AbstractAction() 
                {
                    public void actionPerformed( ActionEvent e )
                    {
                        JFileChooser chooser = new JFileChooser();
                        chooser.setCurrentDirectory(new File(textInfo.getText()));
                        chooser.setDialogType(JFileChooser.OPEN_DIALOG);
                        chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
                        
                        
                        int returnVal = chooser.showOpenDialog(getFrame());
                        if(returnVal == JFileChooser.APPROVE_OPTION) 
                        {
                            String directory = chooser.getSelectedFile().getAbsolutePath();
                            textInfo.setText( directory+"\\" );
                        }
                    }
                }
            );
            getContentPane().add( outputInfoPanel = new JPanel() );
            outputInfoPanel.add( new JLabel( "output Info directory" ) );
            outputInfoPanel.add( textInfo );
            outputInfoPanel.add( browseButtonInfo );
            
            JButton browseButtonOBJ = new JButton( "browse..." );
            browseButtonOBJ.addActionListener( new AbstractAction() 
                {
                    public void actionPerformed( ActionEvent e )
                    {
                        JFileChooser chooser = new JFileChooser();
                        chooser.setCurrentDirectory(new File(textOBJ.getText()));
                        chooser.setDialogType(JFileChooser.OPEN_DIALOG);
                        chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
                        
                        
                        int returnVal = chooser.showOpenDialog(getFrame());
                        if(returnVal == JFileChooser.APPROVE_OPTION) 
                        {
                            String directory = chooser.getSelectedFile().getAbsolutePath();
                            textOBJ.setText( directory+"\\" );
                        }
                    }
                }
            );
            getContentPane().add( outputOBJPanel = new JPanel() );
            outputOBJPanel.add( new JLabel( "output OBJ directory" ) );
            outputOBJPanel.add( textOBJ );
            outputOBJPanel.add( browseButtonOBJ );

            JButton browseButtonEXE = new JButton( "browse..." );
            browseButtonEXE.addActionListener( new AbstractAction() 
                {
                    public void actionPerformed( ActionEvent e )
                    {
                        JFileChooser chooser = new JFileChooser();
                        chooser.setCurrentDirectory(new File(textEXE.getText()));
                        chooser.setDialogType(JFileChooser.OPEN_DIALOG);
                        chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
                        
                        
                        int returnVal = chooser.showOpenDialog(getFrame());
                        if(returnVal == JFileChooser.APPROVE_OPTION) 
                        {
                            String directory = chooser.getSelectedFile().getAbsolutePath();
                            textEXE.setText( directory+"\\" );
                        }
                    }
                }
            );
            getContentPane().add( outputEXEPanel = new JPanel() );
            outputEXEPanel.add( new JLabel( "output EXE directory" ) );
            outputEXEPanel.add( textEXE );
            outputEXEPanel.add( browseButtonEXE );

            JButton okButton = new JButton( "OK" );
            okButton.addActionListener( new AbstractAction() 
                {
                    public void actionPerformed( ActionEvent e )
                    {
                        outputDirectory = text.getText();
                        properties.setProperty( "outputDirectory", outputDirectory );
                        
                        outputInfoDirectory = textInfo.getText();
                        properties.setProperty( "outputInfoDirectory", outputInfoDirectory );
    
                        outputEXEDirectory = textEXE.getText();
                        properties.setProperty( "outputEXEDirectory", outputEXEDirectory );
    
                        outputOBJDirectory = textOBJ.getText();
                        properties.setProperty( "outputOBJDirectory", outputOBJDirectory );
    
                        System.out.println( "directories = " +text.getText() + " " + textInfo.getText() + " " + textEXE.getText() + " " + textOBJ.getText() );
    
                        try
                        {
                            FileOutputStream file = new FileOutputStream( propertiesFile );
                            properties.store( file, "settings configure" );
                            file.close();
                        }
                        catch( IOException e1 ) {}
    
                        hide();
                    }
                }
            );
            JButton cancelButton = new JButton( "CANCEL" );
            cancelButton.addActionListener( new AbstractAction() 
                {
                    public void actionPerformed( ActionEvent e )
                    {
                        hide();
                    }
                }
            );
            getContentPane().add( buttonPanel = new JPanel() );
            buttonPanel.add( okButton );
            buttonPanel.add( cancelButton );

            getContentPane().setLayout( new GridLayout( 0, 1 ) );

            this.setModal( true );
            this.setTitle( "Settings" );
            this.setResizable( false );

            this.setSize( 500, 230 );
        }

        public void show()
        {
            super.show();

            if( outputDirectory != null )
                text.setText( outputDirectory );
            
            if( outputInfoDirectory != null )
                textInfo.setText( outputInfoDirectory );

            if( outputOBJDirectory != null )
                textOBJ.setText( outputEXEDirectory );

            if( outputEXEDirectory != null )
                textEXE.setText( outputEXEDirectory );
        }
    }
}
