/*

    Semant.java - performs type checking....

*/

package Semant;

import Absyn.FieldExpList;
import Absyn.FieldList;
import ErrorMsg.*;
import Temp.Label;
import Translate.Level;
import Types.Type;

import java.util.Stack;


class ControlList
{
	public Object obj=null;
 	public ControlList prec=null,next=null;
 	ControlList() {}
 	ControlList(Object o,ControlList p,ControlList n) {obj=o;prec=p;next=n;}
}

public class Semant
{
    // si usa per dichiarazioni in cui non si eseguono istruzioni
    private static final Translate.Exp dicExp = Translate.Translate.dicExp;
    // si usa come exp restituito in caso di errore semantico
    private static final Translate.Exp errExp = Translate.Translate.errExp;
    private static final ExpTy errExpTy = new ExpTy(errExp, Env._void);

    // the environment
    private Env env;

    private Translate.Translate transAct;
    private Translate.Level levelAct;

    // var for error recovery
    private ErrorMsg errorMsg;

    // var to remember if break ok
    // (later on there will be a reference to a loop necessary to remember what's the last loop to break from)
    private boolean breakOk = false;
    private Stack breakLabelStack = new Stack();


	public Semant(ErrorMsg em, Translate.Level l)
	{
        errorMsg = em;
    	levelAct = l;
 		env = new Env(levelAct);
		transAct=new Translate.Translate(l.frame.wordSize());
	}

	public Semant(ErrorMsg em, Frame.Frame f)
	{
    	errorMsg = em;
	    levelAct = new Translate.Level(f);
 		env = new Env(levelAct);
 		transAct=new Translate.Translate(f.wordSize());
	}


    //method for Exception elimination (error recovery)
    private ExpTy riseError(String mess,int data)
    {
        errorMsg.error(data,mess);
        errorMsg.semanticErrors++;
        return errExpTy;
    }


    // must be called after Semant.transProg(...)
    // return the results of intermediate code
    public Translate.Frag getFragments(){ return Translate.Translate.getResult(); }


    // public method to start type checking
    public ExpTy transProg(Absyn.Exp exp)
    {
        ExpTy result = transExp(exp);
        if(result==null || result.exp==null)
            return null;

        Translate.Exp tte = result.exp;
        if(result.ty.actual() != Env._void)
          tte = transAct.assignExp(new Translate.Ex(new Tree.TEMP(levelAct.frame.RV())), tte);
        transAct.insProgram(tte, levelAct.frame);

        return result;
    }


    // utility functions
    // compare just the references, because each type is represented by one instance of a Types.Type class
    private boolean equalTypes(Types.Type t1, Types.Type t2)
    {
        if(t1 == null || t2 == null) return false;
        // record and nil are "kind of" the same type, thus return true
        if((instanceOfRecord(t1) && t2 == Env._nil) || (t1 == Env._nil && instanceOfRecord(t2))) return true;
        // compare references....
        return (t1.actual() == t2.actual());
    }

    // this is used for mutually recursiv type def's
    private Types.Type lookupType(Symbol.Symbol s, int errpos)
    {
        Types.Type t = ((Types.Type) env.tenv.get(s));
        if(t == null) {
            riseError("Tipo '"+ s.toString()+"' indefinito ", errpos);
            return null;
        }
        return t; // this returns not actual!
    }

    // this is used when it's necessary, that the type is already binded
    private Types.Type lookupBindedType(Symbol.Symbol s, int errpos)
    {
        Types.Type t = ((Types.Type) env.tenv.get(s));
        if(t==null || t instanceof Types.NAME && ((Types.NAME) t).isLoop()) {
            riseError("Tipo '"+s.toString()+"' dichiarato ma non definito ", errpos);
        	return null;
        }
        return t.actual();
    }

    // three simple type-check "shortcuts"
    private boolean instanceOfRecord(Types.Type t)
    {
    	if(t==null) return false;
        return (t.actual() instanceof Types.RECORD);
    }
    private boolean instanceOfArray(Types.Type t)
    {
    	if(t==null) return false;
        return (t.actual() instanceof Types.ARRAY);
    }
    private boolean bothInt(ExpTy op1, ExpTy op2)
    {
        if(op1==null || op2==null) return false;
        Types.Type t1 = op1.ty;
        Types.Type t2 = op2.ty;
	    if(t1==null || t2==null) return false;
        return (equalTypes(t1, Env._int) && equalTypes(t2, Env._int));
    }
    private boolean bothIntOrStr(ExpTy op1, ExpTy op2)
    {
        if(op1==null || op2==null) return false;
        Types.Type t1 = op1.ty;
        Types.Type t2 = op2.ty;
    	if(t1==null || t2==null) return false;
        return ((equalTypes(t1, Env._int) && equalTypes(t2, Env._int)) || (equalTypes(t1, Env._string) && equalTypes(t2, Env._string)));
    }


    //__ dispatch functions

    Types.Type transTy(Absyn.Ty e)
    {
        if(e instanceof Absyn.NameTy) return transTy( (Absyn.NameTy) e );
        if(e instanceof Absyn.RecordTy) return transTy( (Absyn.RecordTy) e );
        if(e instanceof Absyn.ArrayTy) return transTy( (Absyn.ArrayTy) e );
        throw new SemanticError("[errore interno] Unknown subclass of Absyn.Ty in Semant.transTy", e.pos);
    }


    Translate.Exp transDec(Absyn.Dec e)
    {
        if(e instanceof Absyn.FunctionDec) return transDec( (Absyn.FunctionDec) e );
        if(e instanceof Absyn.VarDec) return transDec( (Absyn.VarDec) e );
        if(e instanceof Absyn.TypeDec) return transDec( (Absyn.TypeDec) e );
        throw new SemanticError("[errore interno] Unknown subclass of Absyn.Dec in Semant.transDec", e.pos);
    }

    ExpTy transVar(Absyn.Var e)
    {
        if(e instanceof Absyn.SimpleVar) return transVar( (Absyn.SimpleVar) e );
        if(e instanceof Absyn.FieldVar) return transVar( (Absyn.FieldVar) e );
        if(e instanceof Absyn.SubscriptVar) return transVar( (Absyn.SubscriptVar) e );
        throw new SemanticError("[errore interno] Unknown subclass of Absyn.Var in Semant.transVar", e.pos);
    }

    ExpTy transExp(Absyn.Exp e)
    {
        if(e instanceof Absyn.VarExp) return transExp( (Absyn.VarExp) e );
        if(e instanceof Absyn.NilExp) return transExp( (Absyn.NilExp) e );
        if(e instanceof Absyn.IntExp) return transExp( (Absyn.IntExp) e );
        if(e instanceof Absyn.StringExp) return transExp( (Absyn.StringExp) e );
        if(e instanceof Absyn.CallExp) return transExp( (Absyn.CallExp) e );
        if(e instanceof Absyn.OpExp) return transExp( (Absyn.OpExp) e );
        if(e instanceof Absyn.RecordExp) return transExp( (Absyn.RecordExp) e );
        if(e instanceof Absyn.ArrayExp) return transExp( (Absyn.ArrayExp) e );
        if(e instanceof Absyn.SeqExp) return transExp( (Absyn.SeqExp) e );
        if(e instanceof Absyn.AssignExp) return transExp( (Absyn.AssignExp) e );
        if(e instanceof Absyn.IfExp) return transExp( (Absyn.IfExp) e );
        if(e instanceof Absyn.WhileExp) return transExp( (Absyn.WhileExp) e );
        if(e instanceof Absyn.ForExp) return transExp( (Absyn.ForExp) e );
        if(e instanceof Absyn.BreakExp) return transExp( (Absyn.BreakExp) e );
        if(e instanceof Absyn.LetExp) return transExp( (Absyn.LetExp) e );
        throw new SemanticError("[errore interno] Unknown subclass of Exp in Semant.transExp", e.pos);
    }


    //__ translate functions... for each class of Absyn an overloaded one


    Types.Type transTy(Absyn.NameTy t)
    {
        Types.Type resultT = lookupType(t.name, t.pos);
        // check for use of declared but not defined type (in mutually recursive definitions)
        // (if it is not even declared lookupType would throw an error)
        if(resultT instanceof Types.NAME && ((Types.NAME) resultT).isLoop())  {
            riseError("Tipo '"+t.name+"' dichiarato ma non definito (ciclo nelle definizioni di tipo)", t.pos);
            return null;
        }

        return resultT.actual();
    }

    Types.Type transTy(Absyn.RecordTy t)
    {
        Absyn.FieldList ptr = t.fields;
        Types.RECORD recPtr, rec = new Types.RECORD(null, null, null);

        if(ptr != null) {
            // first element
            rec = new Types.RECORD(ptr.name, lookupType(ptr.typ, ptr.pos) ,null);
            recPtr = rec;
            //further elements
            while(ptr.tail!=null) {
                ptr = ptr.tail;
                recPtr.tail = new Types.RECORD(ptr.name, lookupType(ptr.typ, ptr.pos) ,null);
                recPtr=recPtr.tail;
            }
        }

        return rec;
    }

    Types.Type transTy(Absyn.ArrayTy t)
    {
        Types.Type at = new Types.ARRAY(lookupType(t.typ, t.pos));
        return at;
    }

    Translate.Exp transDec(Absyn.VarDec d)
    {
    	ExpTy initValue = transExp(d.init);
        Types.Type varInitType = initValue.ty.actual();
        Types.Type varDecType = null;

        // if dec specifies a type, it must be the same than the type of the init expression
        // otherwise (if dec does not specify a type), it cannot be void
        if(d.typ != null) {
            varDecType = transTy(d.typ);
            if(!equalTypes(varDecType, varInitType)) {
            	riseError("Dichiarazione di variabile non valida (Il tipo iniziale non corrisponde al tipo attuale) ", d.pos);
            	return dicExp;
            }
        }
        else if(varInitType == Env._void) {
            riseError("L'espressione di inizializzazione per '"+d.name+"' non ha tipo (void).", d.pos);
            return dicExp;
        }
        // if var initialized with nil, the type of var must be specified
        if(varInitType == Env._nil) {
            if(varDecType == null) {
            	riseError("Dichiarazione di variabile non valida (Il tipo della variabile, inizializzata a nil, deve essere specificato) ", d.pos);
            	return dicExp;
            }
            varInitType = varDecType;
        }

        Translate.Access css = levelAct.allocLocal(true);
        env.venv.put(d.name, new VarEntry(css, varInitType ));

        return transAct.assignExp(transAct.simpleVar(css,levelAct),initValue.exp);
    }

    //Cicli nella definizione di tipo
    Translate.Exp transDec(Absyn.TypeDec d)
    {
        Absyn.TypeDec td = d;
        ControlList h = null, cur = null;
        Absyn.TypeDec td2;
        int nDec = 0;

        // Assegnamo agli identificatori di tipo il tipo name
        while(td != null) {
            cur = new ControlList(td,cur,null);
            if(cur.prec == null)
                h = cur;
            else
                cur.prec.next = cur;

            for(td2=td.next; td2!=null; td2=td2.next)
                if(td.name.equals(td2.name))
                    return riseError("Il tipo '"+td.name.toString()+"' e' gia' stato dichiarato nello stesso blocco di mutua ricorsione", td.pos).exp;

            env.tenv.put(td.name, new Types.NAME(td.name));
            td = td.next;
            nDec++;
        }

	    // assegna il binding del tipo ad ogni variabile che ha come binding un'altra variabile
        Types.Type typeP1;
        for(int i=0; i<nDec && h!=null; i++) {
            cur = h;
            for(int j=0; (cur!=null); j++) {
                td = (Absyn.TypeDec)cur.obj;
                if(td.ty instanceof Absyn.NameTy) {
                    typeP1 = lookupType(((Absyn.NameTy)td.ty).name,((Absyn.NameTy)td.ty).pos);
                    if(typeP1!=null) {
                        if(!(typeP1 instanceof Types.NAME && ((Types.NAME)typeP1).isLoop())) {
                            if(cur!=h)
                                cur.prec.next=cur.next;
                            else {
                                h = cur.next;
                                cur.prec=null;
                            }
                            if(cur.next!=null)
                                cur.next.prec=cur.prec;
                            ((Types.NAME)env.tenv.get(td.name)).bind(typeP1);
                        }
                    }
                }
                else {
                    if(cur!=h)
                        cur.prec.next=cur.next;
                    else {
                        h = cur.next;
                        cur.prec=null;
                    }
                    if(cur.next!=null)
                        cur.next.prec=cur.prec;
                    typeP1=transTy(td.ty);
                    ((Types.NAME)env.tenv.get(td.name)).bind(typeP1);
                }
                cur = cur.next;
            }
        }
        if(h!=null) riseError("Alcuni tipi dichiarati formano un ciclo nelle definizioni di tipo", d.pos);

        return dicExp;
    }

    Translate.Exp transDec(Absyn.FunctionDec d)
    {
	    Translate.Level newL;
        // put all function declarations in the environment
        for(Absyn.FunctionDec fptr = d; fptr!=null; fptr=fptr.next)
        {
            for (Absyn.FunctionDec fptr2=fptr.next; fptr2 != null; fptr2=fptr2.next)
            {
                if (fptr.name.equals(fptr2.name))
                    return riseError("La funzione '"+fptr.name.toString()+"' e' gia' stata dichiarata nello stesso blocco di mutua ricorsione", fptr.pos).exp;
            }

            Types.Type result;
            if(fptr.result == null) result = Env._void;
              else result = transTy(fptr.result);

            Absyn.FieldList absynParam = fptr.params;
            Types.RECORD semantParam, sPar;

            if(absynParam==null)
                semantParam = null;
            else {
                // first param
                semantParam =  new Types.RECORD(absynParam.name, lookupBindedType(absynParam.typ, absynParam.pos) ,null);
                sPar = semantParam;

                // further params
                while(absynParam.tail!=null)
                {
                    absynParam=absynParam.tail;

                    sPar.tail = new Types.RECORD(absynParam.name, lookupBindedType(absynParam.typ, absynParam.pos) ,null);
                    sPar = sPar.tail;
                }
            }

            newL=new Translate.Level(levelAct,fptr.name,fieldToBool(fptr.params));

            env.venv.put(fptr.name, new FunEntry(newL,semantParam, result));
        }

        // process bodies
        for(Absyn.FunctionDec fptr = d; fptr!=null; fptr=fptr.next)
        {
            FunEntry fe = (FunEntry) env.venv.get(fptr.name);
            Types.RECORD semantParam = fe.formals, sPar;
            Types.Type result = fe.result;

            env.venv.beginScope();
            env.tenv.beginScope();

            Translate.AccessList om=fe.level.formals;
            for(FieldList fl=fptr.params; fl!=null; fl=fl.tail)
            {
                Types.Type typo = (Type)env.tenv.get(fl.typ);
            	if (typo != null)
                	env.venv.put(fl.name, new VarEntry(om.head, typo));
                else
                    riseError("Errore nella dichiarazione del parametro '"+fl.name+"': tipo "+fl.typ+" sconosciuto", fptr.pos);
                om = om.tail;
            }

            // remember weather already in loop and NOT allow break for this functiondef
            boolean wasOk = breakOk;
            breakOk = false;
            Level temp=levelAct;
            levelAct=fe.level;
            ExpTy functionCode = transExp(fptr.body);
            levelAct=temp;
            if(functionCode==null) continue;
            breakOk = wasOk;

            // type check of return type
            if(!equalTypes(result,functionCode.ty))
            	return riseError("Tipo di ritorno della Definizione di Funzione non valido ("+fptr.name+") ", fptr.pos).exp;

            transAct.functionDec(fe.level, functionCode.exp, (functionCode.ty.actual()!=Env._void));

            env.venv.endScope();
            env.tenv.endScope();

        }

        return dicExp;
    }


    ExpTy transVar(Absyn.SimpleVar v)
    {
        Entry x = (Entry) env.venv.get(v.name);
        if(x==null) return riseError("La variabile '"+v.name.toString()+"' non e' dichiarata ", v.pos);
        if(x instanceof VarEntry) {
            VarEntry ent = (VarEntry) x;
            return new ExpTy(transAct.simpleVar(ent.access,levelAct), ent.ty.actual());
        }
        return riseError("'"+v.name.toString()+"'  una funzione! Non usare come variabile! ", v.pos);
    }


    ExpTy transAssignVar(Absyn.SimpleVar v)
    {
        Entry x = (Entry) env.venv.get(v.name);
        if(x == null)
            return riseError("La variabile '"+v.name.toString()+"' non e' dichiarata ", v.pos);
        if(x instanceof ForVarEntry)
            return riseError(v.name.toString()+" e' una variabile indice, non modificabile", v.pos);
        if(x instanceof VarEntry) {
            VarEntry ent = (VarEntry) x;
            return new ExpTy(transAct.simpleVar(ent.access,levelAct), ent.ty.actual());
        }
        return riseError("'"+v.name.toString()+"'  una funzione! Non usare come variabile! ", v.pos);
    }


    ExpTy transVar(Absyn.FieldVar v)
    {
        ExpTy et = transVar(v.var);
        if(!instanceOfRecord(et.ty))
        	return riseError("Accesso al campo non valido (il valore di sinistra non e' di tipo record) ", v.pos);
        int ii = 0;
        for(Types.RECORD list = (Types.RECORD)et.ty.actual();
            list != null;
            list=list.tail)
      	{
        	if(list.fieldName == v.field)
        		return new ExpTy(transAct.fieldVar(levelAct,et.exp,ii),list.fieldType.actual());
        	ii++;
      	}
      	return riseError("Accesso al campo '"+v.field+"' non valido (campo non trovato nel tipo record) ", v.pos);
    }


    ExpTy transVar(Absyn.SubscriptVar v)
    {
        ExpTy et = transVar(v.var);
        ExpTy index = transExp(v.index);
        if(!instanceOfArray(et.ty))
        	return riseError("Accesso all'array non valido (la variabile non e' un array) ", v.pos);
        //cast sul binding attuale
        Types.ARRAY arr = (Types.ARRAY) et.ty.actual();

        if(!equalTypes(index.ty, Env._int))
        	return riseError("Accesso all'array non valido (l'indice non e' di tipo int) ", v.pos);

        ExpTy result = new ExpTy(transAct.subscriptVar(et.exp,index.exp,levelAct), arr.element);
        return result;
    }

    ExpTy transExp(Absyn.VarExp e)
    {
        return transVar(e.var);
    }

    ExpTy transExp(Absyn.NilExp e)
    {
        return new ExpTy(transAct.intExp(0), Env._nil);
    }


    ExpTy transExp(Absyn.IntExp e)
    {
        return new ExpTy( transAct.intExp(e.value), Env._int);
    }

    ExpTy transExp(Absyn.StringExp e)
    {
        return new ExpTy( transAct.stringExp(levelAct,e.value), Env._string);
    }

    ExpTy transExp(Absyn.CallExp e)
    {
        Entry x = (Entry)env.venv.get(e.func);
        if(x==null)
            return riseError("Funzione '"+e.func.toString()+"' non dichiarata ", e.pos);
        if(!(x instanceof FunEntry))
            return riseError(e.func.toString()+" non e' una funzione ", e.pos);
        FunEntry funEntry = (FunEntry) x;

        Types.RECORD defArgs = funEntry.formals;
        Absyn.ExpList callArgs = e.args;

        Translate.ExpList args = new Translate.ExpList();

        ExpTy errTy = null;
        while(defArgs!=null && callArgs!=null) {
            ExpTy tCallArgs = transExp(callArgs.head);
            if(tCallArgs==null)
                errTy = riseError("Parametro della funzione '"+e.func.toString()+"' non valido", callArgs.head.pos);
            if(!equalTypes(defArgs.fieldType, tCallArgs.ty))
                errTy = riseError("Tipo del parametro della funzione '"+e.func.toString()+"' non valido", callArgs.head.pos);

            args.addInTail(tCallArgs.exp);

            defArgs = defArgs.tail;
            callArgs = callArgs.tail;
        }
        if(defArgs!=null || callArgs!=null)
            errTy = riseError("Numero di argomenti nella chiamata della funzione '"+e.func.toString()+"' non valido ", e.pos);
        if(errTy != null) return errTy;

        Translate.Exp rexp;
        Label funname=((Temp.Label)Frame.Frame.extRoutine.get(e.func.toString()));

        if(funname!=null && funname.equals(((FunEntry)x).level.frame.name))
            rexp = new Translate.Ex(levelAct.frame.externalCall(e.func.toString(),transAct.makeTreeExpList(args)));
        else
            rexp = transAct.callExp(levelAct, ((FunEntry)x).level, args);
        return new ExpTy(rexp, funEntry.result);
    }


    ExpTy transExp(Absyn.OpExp e)
    {
        ExpTy left = transExp(e.left);
        ExpTy right = transExp(e.right);
	    ExpTy ret = null;

        switch(e.oper)
        {
            case Absyn.OpExp.PLUS:
            case Absyn.OpExp.MINUS:
            case Absyn.OpExp.MUL:
            case Absyn.OpExp.DIV:
	            if(!bothInt(left, right)) return riseError("Gli operandi devono essere entrambi di tipo intero ", e.pos);
	            return new ExpTy(transAct.binOpExp(e.oper, left.exp, right.exp), Env._int);
            case Absyn.OpExp.LT:
            case Absyn.OpExp.LE:
            case Absyn.OpExp.GT:
            case Absyn.OpExp.GE:
	            if(!bothIntOrStr(left, right)) return riseError("Gli operandi devono essere entrambi di tipo intero o stringa", e.pos);
	            return new ExpTy(transAct.relOpExp(e.oper, left.exp, right.exp, equalTypes(left.ty, Env._string), levelAct), Env._int);
            case Absyn.OpExp.EQ:
            case Absyn.OpExp.NE:
	            if(left==null || right==null ||
                   !equalTypes(left.ty, right.ty)) return riseError("Tipi non confrontabili per l'operatore '=' ", e.pos);
            	return new ExpTy(transAct.relOpExp(e.oper, left.exp, right.exp, equalTypes(left.ty, Env._string), levelAct), Env._int);
        }

        throw new SemanticError("[errore interno] Semant.transExp(OpExp e)", e.pos);
    }

    ExpTy transExp(Absyn.RecordExp e)
    {
        Types.Type t = lookupBindedType(e.typ, e.pos);

        if(!instanceOfRecord(t)) return riseError("Creazione di record non valida (il tipo non e' un record) ", e.pos);

        Types.RECORD rec = (Types.RECORD) t;
        Absyn.FieldExpList head = e.fields;
        Translate.ExpList initsinv = new Translate.ExpList();
        int fieldnum = fieldNum(rec);
        if(fieldnum!=fieldNum(head))
          return riseError("Numero di campi dichiarati diverso dal numero dei campi assegnati.",e.pos);

        for(Absyn.FieldExpList ptr = head;ptr!=null; ptr=ptr.tail,rec = rec.tail)
        {
            ExpTy init = null;
            if(rec.fieldName.equals(ptr.name)) {
                init = transExp(ptr.init);
                if(!equalTypes(init.ty, rec.fieldType))
                	return riseError("Creazione di record non valida (tipo di inizializzazione del campo '"+rec.fieldName.toString()+"' errato) ", ptr.pos);
                else
                {
                     //Bisogna verificare la compatibilit dell'assegnamento
                    ExpTy ett = transExp(ptr.init);
                    //costruisci la lista di espressioni di inizializzazione invertita
                     initsinv.addInFront(ett.exp);

                     Types.Type tipoass = ett.ty;

                     //Potrebbe accadere in seguito ad un errore nell'espressione di ass.
                     Types.Type tipocampo = rec.fieldType;
                     //Potrebbe accadere in caso di errori prec. nelle dich. dei tipi
                     // In questo caso ritorno null ma potre far ritornare il giusto tipo del record
                     if (tipocampo==Env._void)
                        return riseError("Errore di tipo : Tipo del campo ed assegnamento non compatibili(void).",head.pos);
                     else
                     {
                         // Tutto OK.
                         if (!equalTypes(tipocampo.actual(),tipoass.actual()))
                         //Errore asse. e tipo non compatibili.
                         {
                             return riseError("Errore di tipo : Tipo del campo ed assegnamento non compatibili.",head.pos);
                            // In questo caso ritorno null ma potrei far ritornare il giusto tipo del record
                         }
                     }
                }
            }
            if(init == null)
                return riseError("Creazione di record non valida (campo '"+rec.fieldName.toString()+"' non inizializzato) ", e.pos);
        }


        ExpTy result = new ExpTy(transAct.recordExp(levelAct,fieldnum, initsinv), t);

        return result;
    }

    ExpTy transExp(Absyn.ArrayExp e)
    {
        Types.Type t = lookupBindedType(e.typ, e.pos);

        if(!instanceOfArray(t)) return riseError("Creazione di array non valida (il tipo non e' di tipo array) ", e.pos);
        Types.ARRAY arr = (Types.ARRAY) t;

        ExpTy size = transExp(e.size);
        if(!equalTypes(size.ty, Env._int))
        	return riseError("Creazione di array non valida (la dimensione dell'array deve essere un intero) ", e.pos);

        ExpTy init = transExp(e.init);
        if(!equalTypes(arr.element, init.ty))
        	return riseError("Creazione di array non valida (tipo degli elementi dell'array e tipo dell'inizializzazione non corrispondono) ", e.pos);

        ExpTy result = new ExpTy(transAct.arrayExp(size.exp,init.exp,levelAct), arr);

        return result;
    }

    ExpTy transExp(Absyn.SeqExp e)
    {
        ExpTy et = null;
        Translate.ExpList etl = new Translate.ExpList();
        // crea una lista di exp che contengano istruzioni (dunque non nulle)
        for(Absyn.ExpList el=e.list; el!=null; el=el.tail) {
            et = transExp(el.head);
            if(et!=null && et.exp!=null)
                etl.addInTail(et.exp);
        }

        if(et == null)
            et = new ExpTy(null, Env._void);
        et.exp = transAct.seqExp(etl, et.ty.actual()==Env._void);

        return et;
    }

    ExpTy transExp(Absyn.AssignExp e)
    {
        ExpTy var;
        if(e.var instanceof Absyn.SimpleVar)
            var = transAssignVar((Absyn.SimpleVar)e.var);
        else
            var = transVar(e.var);
        ExpTy exp = transExp(e.exp);

        Types.Type sx=var.ty.actual();
      	Types.Type dx=exp.ty.actual();

        if(!equalTypes(sx, dx))
        	if(dx != Env._nil)
        		return riseError("Assegnamento non valido (tipi non corrispondenti) ", e.pos);
        	else
        		if(!(sx instanceof Types.RECORD)) return riseError("Errore di tipo : nil solo a tipi record.",e.pos);

        ExpTy result = new ExpTy( transAct.assignExp(var.exp,exp.exp), Env._void);

        return result;

    }


    ExpTy transExp(Absyn.IfExp e)
    {
        ExpTy test = transExp(e.test);
        if(!equalTypes(test.ty, Env._int)) riseError("Espressione di test deve essere di tipo int", e.pos);
        ExpTy thenclause = transExp(e.thenclause);
        ExpTy elseclause = e.elseclause!=null ? transExp(e.elseclause)
                                              : new ExpTy(null, Env._void);
        if(!equalTypes(thenclause.ty, elseclause.ty))
            return riseError("If-Then-Else non valido (i rami dell'if non sono dello stesso tipo)", e.pos);

        Translate.Exp appTT = errExp;
        if(test.exp != null)
            appTT = transAct.ifThenElseExp(test.exp,thenclause.exp,elseclause.exp);
        if(!equalTypes(thenclause.ty,Env._void))
        	appTT=new Translate.Ex(appTT.unEx());
        else
            appTT=new Translate.Nx(appTT.unNx());

        return new ExpTy(appTT, thenclause.ty);
    }

    ExpTy transExp(Absyn.WhileExp e)
    {
    	breakLabelStack.push(new Temp.Label());
        ExpTy test = transExp(e.test);

        // remember weather already in loop and allow break for this loop
        boolean wasOk = breakOk;
        breakOk = true;
        ExpTy body = transExp(e.body);
        breakOk = wasOk;

        if(!equalTypes(test.ty, Env._int))
        	riseError("Espressione di test deve essere di tipo int", e.pos);
        if(!equalTypes(body.ty, Env._void))
        	riseError("Corpo dell'espressione del 'while' non deve ritornare un valore", e.pos);

        ExpTy result = new ExpTy(transAct.whileExp(test.exp,body.exp,(Temp.Label)breakLabelStack.peek()),Env._void);
        breakLabelStack.pop();

        return result;
    }

    ExpTy transExp(Absyn.ForExp e)
    {
    	breakLabelStack.push(new Temp.Label());
		// remember weather already in loop and allow break for this loop
        boolean wasOk = breakOk;
        breakOk = true;

    	ExpTy init = transExp(e.var.init);
        if(!equalTypes(init.ty,Env._int))
        	riseError("Variabile indice non valida (Il tipo non e' intero) ", e.var.pos);

        ExpTy hi = transExp(e.hi);
        if(!equalTypes(hi.ty, Env._int))
        	riseError("Espressione 'to' deve essere di tipo int", e.hi.pos);

        env.venv.beginScope();
        env.tenv.beginScope();
        Translate.Access iacc=levelAct.allocLocal(true);
        env.venv.put(e.var.name, new ForVarEntry(iacc, Env._int ));

        ExpTy body = transExp(e.body);

        env.venv.endScope();
     	env.tenv.endScope();

        breakOk = wasOk;
        if(!equalTypes(body.ty, Env._void))
            return riseError("Corpo dell'espressione 'for' non deve ritornare un valore", e.pos);

        ExpTy result = new ExpTy(transAct.forExp(transAct.simpleVar(iacc,levelAct),init.exp,hi.exp,body.exp,
                                                (Temp.Label)breakLabelStack.peek(), levelAct), Env._void);
        breakLabelStack.pop();

        return result;
    }

    ExpTy transExp(Absyn.BreakExp e)
    {
        if(!breakOk) return riseError("Break deve essere nel corpo del ciclo while", e.pos);
        return new ExpTy( transAct.breakExp((Temp.Label)breakLabelStack.peek()), Env._void);
    }


    ExpTy transExp(Absyn.LetExp e)
    {
        env.venv.beginScope();
        env.tenv.beginScope();
        Translate.ExpList decs = new Translate.ExpList();

        // translate decs if specified
        if(e.decs!=null) {
           for(Absyn.DecList p=e.decs; p!=null; p=p.tail) {
			 // translate head dec
             Translate.Exp pR = transDec(p.head);
             if(pR != null) decs.addInTail(pR);
           }
        }

        // if someting between in and end
        ExpTy body = e.body!=null ? transExp(e.body)
                                  : new ExpTy(null, Env._void);

        env.venv.endScope();
        env.tenv.endScope();

        if(body == null) return errExpTy;
        return new ExpTy(transAct.letExp(decs,body.exp,equalTypes(body.ty, Env._void)), body.ty);
    }

    Util.BoolList fieldToBool(Absyn.FieldList e)
    {
    	if(e == null) return null;
    	return new Util.BoolList(true,fieldToBool(e.tail));
    }

	private int fieldNum(FieldExpList head)
	{
		return head==null ? 0
                          : 1 + fieldNum(head.tail);
	}

    private int fieldNum(Types.RECORD r)
    {
	   if(r == null || r.fieldName == null) return 0;
	   return 1 + fieldNum(r.tail);
    }

    public void printIC(java.io.PrintStream out)
    {
        Tree.Print treePrinter = new Tree.Print(out);
        treePrinter.sayln("\n\n---- Intermediate Code Tree");
        boolean lastData = false;
        for(Translate.Frag f=getFragments(); f!=null; f=f.next) {
          if(f instanceof Translate.ProcFrag) {
            treePrinter.sayln("\n-- fragment '"+((Translate.ProcFrag)f).frame.debugName+"'");
            treePrinter.prStm(((Translate.ProcFrag)f).body);
            lastData = false;
          }
          if(f instanceof Translate.DataFrag) {
            if(lastData)
              treePrinter.say(((Translate.DataFrag)f).data);
            else
              treePrinter.sayln("\n" + ((Translate.DataFrag)f).data);
            lastData = true;
          }
        }
    }

    public void printCanonIC(java.io.PrintStream out)
    {
        Tree.Print treePrinter = new Tree.Print(out);
        treePrinter.sayln("\n\n\n---- Canonical Code Tree");
        boolean lastData = false;
        for(Translate.Frag f=getFragments(); f!=null; f=f.next) {
          if(f instanceof Translate.ProcFrag) {
            treePrinter.sayln("\n-- canonical fragment '"+((Translate.ProcFrag)f).frame.debugName+"'");
            treePrinter.prStmList(((Translate.ProcFrag)f).cbody);
            lastData = false;
          }
          if(f instanceof Translate.DataFrag) {
              if(lastData)
                treePrinter.say(((Translate.DataFrag)f).data);
              else
                treePrinter.sayln("\n" + ((Translate.DataFrag)f).data);
             lastData = true;
          }
        }
    }
}
